import glob
import os
import re
import sys
import warnings
from enum import auto
from pathlib import Path
from typing import Any, Literal, Optional, Union, cast

import inifix
import numpy as np

from nonos.api._angle_parsing import _parse_planet_file
from nonos.logging import logger

if sys.version_info >= (3, 11):
    from enum import StrEnum
    from typing import assert_never
else:
    from typing_extensions import assert_never

    from nonos._backports import StrEnum


class Code(StrEnum):
    IDEFIX = auto()
    PLUTO = auto()
    FARGO3D = auto()
    FARGO_ADSG = auto()


_INIFILES_LOOKUP_TABLE = {
    "idefix.ini": Code.IDEFIX,
    "pluto.ini": Code.PLUTO,
    "variables.par": Code.FARGO3D,
}


class Parameters:
    def __init__(
        self,
        *,
        inifile: Optional[str] = None,
        code: Optional[str] = None,
        directory: Optional[str] = None,
    ) -> None:
        warnings.warn(
            "The Parameters class is deprecated "
            "and will be removed in a future version. "
            "Instead, use nonos.loaders.loader_from ",
            category=DeprecationWarning,
            stacklevel=2,
        )
        if not directory:
            directory = os.getcwd()
        elif not os.path.exists(directory):
            raise FileNotFoundError(f"No such file or directory {directory}")
        elif not os.path.isdir(directory):
            raise ValueError(f"{directory} is not a directory")

        self.directory = directory
        self.paramfile = inifile or ""
        self.code: Code
        if code:
            self.code = Code(code)

        if code and self.paramfile:
            found = Path(self.directory).joinpath(self.paramfile).is_file()
            if not found:
                raise FileNotFoundError(f"{self.paramfile} not found.")
        elif not code and not self.paramfile:
            found_dict = {
                paramfile: Path(self.directory).joinpath(paramfile).is_file()
                for paramfile in _INIFILES_LOOKUP_TABLE
            }
            nfound = sum(list(found_dict.values()))
            if nfound == 0:
                raise FileNotFoundError(
                    f"{', '.join(_INIFILES_LOOKUP_TABLE.keys())} not found."
                )
            elif nfound > 1:
                raise RuntimeError("found more than one possible ini file.")
            self.paramfile = list(_INIFILES_LOOKUP_TABLE.keys())[
                list(found_dict.values()).index(True)
            ]
            self.code = _INIFILES_LOOKUP_TABLE[self.paramfile]
        elif code or self.paramfile:
            raise ValueError("both inifile and code have to be given.")

    def loadIniFile(self) -> None:
        FrameT = Literal["F", "C", None]

        # inifix.load uses dynamic type inference and doesn't provide type safety at
        # all. Here we define minimal classes that take in arbitrary keyword arguments
        # and store selected values as attributes in a typechecker-friendly way. This
        # implements type safety for the few parameters that we actually care about at
        # typecheck time *and* at runtime.

        class IdefixIniOutput:
            def __init__(self, *, vtk, **_kwargs) -> None:
                self.vtk = float(vtk)

        class IdefixIniHydro:
            def __init__(self, **kwargs) -> None:
                self.rotation: Optional[float]
                if "rotation" in kwargs:
                    self.rotation = float(kwargs["rotation"])
                else:
                    self.rotation = None

        class IdefixIni:
            def __init__(self, *, Hydro, Output, **_kwargs) -> None:
                self.hydro = IdefixIniHydro(**Hydro)
                self.output = IdefixIniOutput(**Output)

        class PlutoIniOutput:
            def __init__(self, *, vtk, **_kwargs) -> None:
                self.vtk = int(list(vtk)[0])

        class PlutoIni:
            def __init__(self, **kwargs) -> None:
                self.output = PlutoIniOutput(**kwargs["Static Grid Output"])

        class Fargo3DIni:
            def __init__(self, *, NINTERM, DT, FRAME, **kwargs) -> None:
                self.NINTERM = int(NINTERM)
                self.DT = float(DT)
                self.FRAME: FrameT
                self.OMEGAFRAME: Optional[float]

                if (str_frame := str(FRAME)) in ("F", "C"):
                    str_frame = cast(Literal["F", "C"], str_frame)
                    self.FRAME = str_frame
                else:
                    self.FRAME = None

                if self.FRAME == "F":
                    self.OMEGAFRAME = float(kwargs["OMEGAFRAME"])
                else:
                    self.OMEGAFRAME = None

        class FargoADSGIni:
            def __init__(self, *, Ninterm, DT, Frame, **kwargs) -> None:
                self.Ninterm = int(Ninterm)
                self.DT = float(DT)
                self.Frame: FrameT
                self.OmegaFrame: Optional[float]

                if (str_frame := str(Frame)) in ("F", "C"):
                    str_frame = cast(Literal["F", "C"], str_frame)
                    self.Frame = str_frame
                else:
                    self.Frame = None

                if self.Frame == "F":
                    self.OmegaFrame = float(kwargs["OmegaFrame"])
                else:
                    self.OmegaFrame = None

        self.vtk: float
        self.omegaframe: Optional[float]
        self.frame: FrameT
        self.inifile = inifix.load(os.path.join(self.directory, self.paramfile))
        if self.code is Code.IDEFIX:
            idefix_ini = IdefixIni(**self.inifile)
            self.vtk = idefix_ini.output.vtk
            self.omegaframe = idefix_ini.hydro.rotation
            if self.omegaframe is None:
                self.frame = None
            else:
                self.frame = "F"
        elif self.code is Code.PLUTO:
            pluto_ini = PlutoIni(**self.inifile)
            self.vtk = pluto_ini.output.vtk
            self.omegaframe = None
            self.frame = None
        elif self.code is Code.FARGO3D:
            fargo3D_ini = Fargo3DIni(**self.inifile)
            self.vtk = fargo3D_ini.NINTERM * fargo3D_ini.DT
            if fargo3D_ini.FRAME == "F":
                self.omegaframe = fargo3D_ini.OMEGAFRAME
                self.frame = "F"
            else:
                self.omegaframe = None
                self.frame = None
        elif self.code is Code.FARGO_ADSG:
            fargoADSG_ini = FargoADSGIni(**self.inifile)
            self.vtk = fargoADSG_ini.Ninterm * fargoADSG_ini.DT
            if fargoADSG_ini.Frame == "F":
                self.omegaframe = fargoADSG_ini.OmegaFrame
                self.frame = fargoADSG_ini.Frame
            else:
                self.omegaframe = None
                self.frame = None
        else:
            assert_never(self.code)

    def loadPlanetFile(
        self, *, planet_number: Optional[int] = None, planet_file: Optional[str] = None
    ) -> None:
        planet_file = _parse_planet_file(
            planet_number=planet_number, planet_file=planet_file
        )
        del planet_number

        if self.code is Code.IDEFIX or self.code is Code.FARGO3D:
            if Path(self.directory).joinpath(planet_file).is_file():
                columns = np.loadtxt(os.path.join(self.directory, planet_file)).T
                self.qpl = columns[7]
                self.dpl = np.sqrt(np.sum(columns[1:4] ** 2, axis=0))
                if self.code is Code.FARGO3D and self.inifile["FRAME"] == "C":
                    self.omegaframe = np.sqrt((1.0 + self.qpl) / pow(self.dpl, 3.0))
                    self.frame = "C"
                # TODO: change (x,y,z) and (vx,vy,vz) when rotation
                self.dtpl = columns[0]
                self.xpl = columns[1]
                self.ypl = columns[2]
                self.zpl = columns[3]
                self.vxpl = columns[4]
                self.vypl = columns[5]
                self.vzpl = columns[6]
                self.tpl = columns[8]
            else:
                raise FileNotFoundError(f"{planet_file} not found")
        elif self.code is Code.FARGO_ADSG:
            if Path(self.directory).joinpath(planet_file).is_file():
                columns = np.loadtxt(os.path.join(self.directory, planet_file)).T
                self.qpl = columns[5]
                self.dpl = np.sqrt(np.sum(columns[1:3] ** 2, axis=0))
                if self.inifile["Frame"] == "C":
                    self.omegaframe = np.sqrt((1.0 + self.qpl) / pow(self.dpl, 3.0))
                    self.frame = "C"
                # TODO: change (x,y,z) and (vx,vy,vz) when rotation
                self.dtpl = columns[0]
                self.xpl = columns[1]
                self.ypl = columns[2]
                self.zpl = 0.0
                self.vxpl = columns[3]
                self.vypl = columns[4]
                self.vzpl = 0.0
                self.tpl = columns[7]
            else:
                raise FileNotFoundError(f"{planet_file} not found")
        else:
            raise NotImplementedError(
                f"{planet_file} not found for {self.code}. For now, you can't rotate the grid with the planet."
            )

        if self.omegaframe is None or self.frame == "C":
            hx = self.ypl * self.vzpl - self.zpl * self.vypl
            hy = self.zpl * self.vxpl - self.xpl * self.vzpl
            hz = self.xpl * self.vypl - self.ypl * self.vxpl
            hhor = np.sqrt(hx * hx + hy * hy)

            h2 = hx * hx + hy * hy + hz * hz
            h = np.sqrt(h2)
            self.ipl = np.arcsin(hhor / h)

            Ax = (
                self.vypl * hz - self.vzpl * hy - (1.0 + self.qpl) * self.xpl / self.dpl
            )
            Ay = (
                self.vzpl * hx - self.vxpl * hz - (1.0 + self.qpl) * self.ypl / self.dpl
            )
            Az = (
                self.vxpl * hy - self.vypl * hx - (1.0 + self.qpl) * self.zpl / self.dpl
            )

            self.epl = np.sqrt(Ax * Ax + Ay * Ay + Az * Az) / (1.0 + self.qpl)
            self.apl = h * h / ((1.0 + self.qpl) * (1.0 - self.epl * self.epl))
            # else:
            #     raise NotImplementedError(
            #         "We do not yet compute eccentricity, inclination and semi-major axis if fixed rotating frame."
            #     )

    def countSimuFiles(self) -> None:
        if self.code is Code.FARGO3D or self.code is Code.FARGO_ADSG:
            self.data_files = [
                fn
                for fn in glob.glob1(self.directory, "gasdens*.dat")
                if re.match(r"gasdens\d+.dat", fn)
            ]
        elif self.code is Code.IDEFIX or self.code is Code.PLUTO:
            self.data_files = list(glob.glob1(self.directory, "data.*.vtk"))
        else:
            assert_never(self.code)

    def loadSimuFile(
        self,
        input_dataset: Union[int, str],
        /,
        *,
        geometry: str = "unknown",
        cell: str = "edges",
        fluid: Optional[str] = None,
    ) -> "DataStructure":
        if fluid is not None and self.code != Code.FARGO3D:
            raise ValueError("fluid is defined only for fargo3d outputs")
        output_number, filename = funnel_on_type(
            input_dataset, code=self.code, directory=self.directory
        )
        self.on = output_number
        codeReadFormat = CodeReadFormat()
        if self.code is Code.FARGO3D:
            return codeReadFormat.fargo3dReadDat(
                self.on, directory=self.directory, inifile=self.paramfile, fluid=fluid
            )
        elif self.code is Code.FARGO_ADSG:
            return codeReadFormat.fargoAdsgReadDat(
                self.on, directory=self.directory
            )  # , inifile=self.paramfile)
        elif self.code is Code.IDEFIX or self.code is Code.PLUTO:
            return codeReadFormat.idfxReadVTK(filename, geometry=geometry, cell=cell)
        else:
            assert_never(self.code)


def funnel_on_type(
    input_dataset: Union[int, str], /, *, code: str, directory="."
) -> tuple[int, str]:
    _code = Code(code)
    if _code is Code.FARGO3D or _code is Code.FARGO_ADSG:
        if isinstance(input_dataset, str):
            raise TypeError(f"on can only be an int for {code}")
        return input_dataset, ""
    elif _code is Code.IDEFIX or _code is Code.PLUTO:
        if isinstance(input_dataset, str):
            filename = os.path.join(directory, input_dataset)
            if (m := re.search(r"\d+", input_dataset)) is None:
                raise ValueError("filename format is not correct")
            else:
                on = int(m.group())
        elif isinstance(input_dataset, (int, np.integer)):
            on = input_dataset
            filename = os.path.join(directory, f"data.{on:04d}.vtk")
        else:
            raise TypeError(
                f"input_dataset type ({type(input_dataset)}) not recognized (should be int or str)"
            )
        return (on, filename)
    else:
        assert_never(_code)


class DataStructure:
    """
    Class that helps create the datastructure
    in the idfxReadVTK and fargo3dReadDat functions
    """

    __slots__ = (
        "data",
        "geometry",
        "t",
        "periodicity",
        "n1",
        "n2",
        "n3",
        "x1",
        "x2",
        "x3",
    )
    data: dict[str, Any]
    geometry: str
    t: np.ndarray
    periodicity: tuple[bool, bool, bool]
    n1: int
    n2: int
    n3: int
    x1: "np.ndarray[Any, np.dtype[np.float32 | np.float64]]"
    x2: "np.ndarray[Any, np.dtype[np.float32 | np.float64]]"
    x3: "np.ndarray[Any, np.dtype[np.float32 | np.float64]]"


class CodeReadFormat:
    def __init__(self):
        warnings.warn(
            "The CodeReadFormat class is deprecated "
            "and will be removed in a future version. "
            "It is replaced internally with nonos._readers.binary, "
            "which isn't intended for external use.",
            category=DeprecationWarning,
            stacklevel=2,
        )

    def idfxReadVTK(
        self, filename, *, geometry="unknown", cell="edges", computedata=True
    ) -> DataStructure:
        """
        Adapted from Geoffroy Lesur
        Function that reads a vtk file in polar coordinates
        """
        nfound = len(glob.glob(filename))
        if nfound != 1:
            raise FileNotFoundError("In idfxReadVTK: %s not found." % filename)

        fid = open(filename, "rb")

        # define our datastructure
        V = DataStructure()

        # raw data which will be read from the file
        V.data = {}

        # initialize geometry
        if geometry not in ("unknown", "cartesian", "polar", "spherical"):
            raise ValueError(f"Unknown geometry value: {geometry!r}")
        V.geometry = geometry

        # datatype we read
        dt = np.dtype(">f")  # Big endian single precision floats
        dint = np.dtype(">i4")  # Big endian integer

        s = fid.readline()  # VTK DataFile Version x.x
        s = fid.readline()  # Comments

        s = fid.readline()  # BINARY
        s = fid.readline()  # DATASET RECTILINEAR_GRID or STRUCTURED_GRID
        slist = s.split()

        s = fid.readline()  # DIMENSIONS NX NY NZ
        slist = s.split()
        entry = str(slist[0], "utf-8")
        if entry == "FIELD":
            nfield = int(slist[2])
            for _field in range(nfield):
                s = fid.readline()
                slist = s.split()
                entry = str(slist[0], "utf-8")
                if entry == "TIME":
                    V.t = np.fromfile(fid, dt, 1)
                elif entry == "GEOMETRY":
                    g = np.fromfile(fid, dint, 1)
                    if g == 0:
                        thisgeometry = "cartesian"
                    elif g == 1:
                        thisgeometry = "polar"
                    elif g == 2:
                        thisgeometry = "spherical"
                    elif g == 3:
                        thisgeometry = "cylindrical"
                    else:
                        fid.close()
                        raise ValueError(
                            f"Unknown value for GEOMETRY flag ('{g}') was found in the VTK file."
                        )

                    if V.geometry != "unknown":
                        # We already have a proposed geometry, check that what is read from the file matches
                        if thisgeometry != V.geometry:
                            fid.close()
                            raise ValueError(
                                f"geometry argument ('{V.geometry}') is inconsistent with GEOMETRY flag from the VTK file ('{thisgeometry}')"
                            )
                    V.geometry = thisgeometry
                elif entry == "PERIODICITY":
                    pint = np.fromfile(fid, dint, 3)
                    V.periodicity = (bool(pint[0]), bool(pint[1]), bool(pint[2]))
                else:
                    fid.close()
                    raise ValueError(f"Received unknown field: '{entry}'.")

                s = fid.readline()  # extra linefeed

            # finished reading the field entry
            # read next line
            s = fid.readline()  # DIMENSIONS...

        if V.geometry == "unknown":
            fid.close()
            raise RuntimeError(
                "Geometry couldn't be determined from data. "
                "Try to set the geometry keyword argument explicitly."
            )

        slist = s.split()  # DIMENSIONS....
        V.n1 = int(slist[1])
        V.n2 = int(slist[2])
        V.n3 = int(slist[3])

        z: Union["np.ndarray", "np.memmap"]

        if V.geometry == "cartesian":
            # CARTESIAN geometry
            s = fid.readline()  # X_COORDINATES NX float
            inipos = (
                fid.tell()
            )  # we store the file pointer position before computing points
            logger.debug("loading the X-grid cells: {}", V.n1)
            x = np.memmap(
                fid, mode="r", dtype=dt, offset=inipos, shape=V.n1
            )  # some smart memory efficient way to store the array
            newpos = (
                dt.itemsize * 1 * V.n1 + inipos
            )  # we calculate the offset that we would expect normally with a np.fromfile
            fid.seek(
                newpos, os.SEEK_SET
            )  # we set the file pointer position to this offset
            s = fid.readline()  # Extra line feed added by idefix

            s = fid.readline()  # Y_COORDINATES NY float
            inipos = (
                fid.tell()
            )  # we store the file pointer position before computing points
            logger.debug("loading the Y-grid cells: {}", V.n2)
            y = np.memmap(
                fid, mode="r", dtype=dt, offset=inipos, shape=V.n2
            )  # some smart memory efficient way to store the array
            newpos = (
                dt.itemsize * 1 * V.n2 + inipos
            )  # we calculate the offset that we would expect normally with a np.fromfile
            fid.seek(
                newpos, os.SEEK_SET
            )  # we set the file pointer position to this offset
            s = fid.readline()  # Extra line feed added by idefix

            s = fid.readline()  # Z_COORDINATES NZ float
            inipos = (
                fid.tell()
            )  # we store the file pointer position before computing points
            logger.debug("loading the Z-grid cells: {}", V.n3)
            z = np.memmap(
                fid, mode="r", dtype=dt, offset=inipos, shape=V.n3
            )  # some smart memory efficient way to store the array
            newpos = (
                dt.itemsize * 1 * V.n3 + inipos
            )  # we calculate the offset that we would expect normally with a np.fromfile
            fid.seek(
                newpos, os.SEEK_SET
            )  # we set the file pointer position to this offset
            s = fid.readline()  # Extra line feed added by idefix

            s = fid.readline()  # POINT_DATA NXNYNZ

            slist = s.split()
            point_type = str(slist[0], "utf-8")
            npoints = int(slist[1])
            s = fid.readline()  # EXTRA LINE FEED

            if point_type == "CELL_DATA" or cell == "centers":
                # The file contains face coordinates, so we extrapolate to get the cell center coordinates.
                if V.n1 > 1:
                    V.n1 = V.n1 - 1
                    if cell == "centers":
                        V.x1 = 0.5 * (x[1:] + x[:-1])
                    elif cell == "edges":
                        V.x1 = x
                else:
                    V.x1 = x
                if V.n2 > 1:
                    V.n2 = V.n2 - 1
                    if cell == "centers":
                        V.x2 = 0.5 * (y[1:] + y[:-1])
                    elif cell == "edges":
                        V.x2 = y
                else:
                    V.x2 = y
                if V.n3 > 1:
                    V.n3 = V.n3 - 1
                    if cell == "centers":
                        V.x3 = 0.5 * (z[1:] + z[:-1])
                    elif cell == "edges":
                        V.x3 = z
                else:
                    V.x3 = z
            elif point_type == "POINT_DATA":
                V.x1 = x
                V.x2 = y
                V.x3 = z

            if V.n1 * V.n2 * V.n3 != npoints:
                fid.close()
                raise ValueError(
                    "In idfxReadVTK: Grid size (%d) incompatible with number of points (%d) in the data set"
                    % (V.n1 * V.n2 * V.n3, npoints)
                )

        else:
            # POLAR or SPHERICAL coordinates
            if V.n3 == 1:
                is2d = 1
            else:
                is2d = 0

            s = fid.readline()  # POINTS NXNYNZ float
            slist = s.split()
            npoints = int(slist[1])

            inipos = (
                fid.tell()
            )  # we store the file pointer position before computing points
            # print(inipos)
            # points = np.fromfile(fid, dt, 3 * npoints)
            logger.debug("loading grid (shape = ({},{},{}))", V.n1, V.n2, V.n3)
            points = np.memmap(
                fid, mode="r", dtype=dt, offset=inipos, shape=3 * npoints
            )  # some smart memory efficient way to store the array
            # print(fid.tell())
            newpos = (
                dt.itemsize * 3 * npoints + inipos
            )  # we calculate the offset that we would expect normally with a np.fromfile
            fid.seek(
                newpos, os.SEEK_SET
            )  # we set the file pointer position to this offset
            # print(fid.tell())
            s = fid.readline()  # EXTRA LINE FEED

            # V.points=points
            if V.n1 * V.n2 * V.n3 != npoints:
                fid.close()
                raise ValueError(
                    "In idfxReadVTK: Grid size (%d) incompatible with number of points (%d) in the data set"
                    % (V.n1 * V.n2 * V.n3, npoints)
                )

            # Reconstruct the polar coordinate system
            x1d = points[::3]
            y1d = points[1::3]
            z1d = points[2::3]

            xcart = np.transpose(x1d.reshape(V.n3, V.n2, V.n1))
            ycart = np.transpose(y1d.reshape(V.n3, V.n2, V.n1))
            zcart = np.transpose(z1d.reshape(V.n3, V.n2, V.n1))

            # Reconstruct the polar coordinate system
            if V.geometry == "polar":
                r = np.sqrt(xcart[:, 0, 0] ** 2 + ycart[:, 0, 0] ** 2)
                theta = np.unwrap(np.arctan2(ycart[0, :, 0], xcart[0, :, 0]))
                z = zcart[0, 0, :]

                s = fid.readline()  # CELL_DATA (NX-1)(NY-1)(NZ-1)
                slist = s.split()
                data_type = str(slist[0], "utf-8")
                if data_type != "CELL_DATA":
                    fid.close()
                    raise ValueError(
                        "In idfxReadVTK: this routine expect 'CELL DATA' as produced by PLUTO, not '%s'."
                        % data_type
                    )
                s = fid.readline()  # Line feed

                if cell == "edges":
                    if V.n1 > 1:
                        V.n1 = V.n1 - 1
                        V.x1 = r
                    else:
                        V.x1 = r
                    if V.n2 > 1:
                        V.n2 = V.n2 - 1
                        V.x2 = theta
                    else:
                        V.x2 = theta
                    if V.n3 > 1:
                        V.n3 = V.n3 - 1
                        V.x3 = z
                    else:
                        V.x3 = z

                # Perform averaging on coordinate system to get cell centers
                # The file contains face coordinates, so we extrapolate to get the cell center coordinates.
                elif cell == "centers":
                    if V.n1 > 1:
                        V.n1 = V.n1 - 1
                        V.x1 = 0.5 * (r[1:] + r[:-1])
                    else:
                        V.x1 = r
                    if V.n2 > 1:
                        V.n2 = V.n2 - 1
                        V.x2 = (0.5 * (theta[1:] + theta[:-1]) + np.pi) % (
                            2.0 * np.pi
                        ) - np.pi
                    else:
                        V.x2 = theta
                    if V.n3 > 1:
                        V.n3 = V.n3 - 1
                        V.x3 = 0.5 * (z[1:] + z[:-1])
                    else:
                        V.x3 = z

            # Reconstruct the spherical coordinate system
            if V.geometry == "spherical":
                if is2d:
                    r = np.sqrt(xcart[:, 0, 0] ** 2 + ycart[:, 0, 0] ** 2)
                    phi = np.unwrap(
                        np.arctan2(zcart[0, V.n2 // 2, :], xcart[0, V.n2 // 2, :])
                    )
                    theta = np.arccos(
                        ycart[0, :, 0]
                        / np.sqrt(xcart[0, :, 0] ** 2 + ycart[0, :, 0] ** 2)
                    )
                else:
                    r = np.sqrt(
                        xcart[:, 0, 0] ** 2 + ycart[:, 0, 0] ** 2 + zcart[:, 0, 0] ** 2
                    )
                    phi = np.unwrap(
                        np.arctan2(
                            ycart[V.n1 // 2, V.n2 // 2, :],
                            xcart[V.n1 // 2, V.n2 // 2, :],
                        )
                    )
                    theta = np.arccos(
                        zcart[0, :, 0]
                        / np.sqrt(
                            xcart[0, :, 0] ** 2
                            + ycart[0, :, 0] ** 2
                            + zcart[0, :, 0] ** 2
                        )
                    )

                s = fid.readline()  # CELL_DATA (NX-1)(NY-1)(NZ-1)
                slist = s.split()
                data_type = str(slist[0], "utf-8")
                if data_type != "CELL_DATA":
                    fid.close()
                    raise ValueError(
                        "In idfxReadVTK: this routine expect 'CELL DATA' as produced by PLUTO, not '%s'."
                        % data_type
                    )
                s = fid.readline()  # Line feed

                if cell == "edges":
                    if V.n1 > 1:
                        V.n1 = V.n1 - 1
                        V.x1 = r
                    else:
                        V.x1 = r
                    if V.n2 > 1:
                        V.n2 = V.n2 - 1
                        V.x2 = theta
                    else:
                        V.x2 = theta
                    if V.n3 > 1:
                        V.n3 = V.n3 - 1
                        V.x3 = phi
                    else:
                        V.x3 = phi

                # Perform averaging on coordinate system to get cell centers
                # The file contains face coordinates, so we extrapolate to get the cell center coordinates.
                elif cell == "centers":
                    if V.n1 > 1:
                        V.n1 = V.n1 - 1
                        V.x1 = 0.5 * (r[1:] + r[:-1])
                    else:
                        V.x1 = r
                    if V.n2 > 1:
                        V.n2 = V.n2 - 1
                        V.x2 = 0.5 * (theta[1:] + theta[:-1])
                    else:
                        V.x2 = theta
                    if V.n3 > 1:
                        V.n3 = V.n3 - 1
                        V.x3 = 0.5 * (phi[1:] + phi[:-1])
                    else:
                        V.x3 = phi

        if computedata:
            logger.debug("loading data arrays")
            while 1:
                s = fid.readline()  # SCALARS/VECTORS name data_type (ex: SCALARS imagedata unsigned_char)
                # print repr(s)
                if len(s) < 2:  # leave if end of file
                    break
                slist = s.split()
                datatype = str(slist[0], "utf-8")
                varname = str(slist[1], "utf-8").upper()
                if datatype == "SCALARS":
                    fid.readline()  # LOOKUP TABLE

                    inipos = (
                        fid.tell()
                    )  # we store the file pointer position before computing points
                    # array = np.fromfile(fid, dt, V.n1 * V.n2 * V.n3).reshape(V.n3, V.n2, V.n1)
                    array = np.memmap(
                        fid, mode="r", dtype=dt, offset=inipos, shape=V.n1 * V.n2 * V.n3
                    ).reshape(
                        V.n3, V.n2, V.n1
                    )  # some smart memory efficient way to store the array
                    newpos = (
                        dt.itemsize * V.n1 * V.n2 * V.n3 + inipos
                    )  # we calculate the offset that we would expect normally with a np.fromfile
                    fid.seek(
                        newpos, os.SEEK_SET
                    )  # we set the file pointer position to this offset

                    V.data[varname] = np.transpose(array)
                elif datatype == "VECTORS":
                    inipos = (
                        fid.tell()
                    )  # we store the file pointer position before computing points
                    nelements = 3 * V.n1 * V.n2 * V.n3
                    Q = np.memmap(
                        fid, mode="r", dtype=dt, offset=inipos, shape=nelements
                    )  # some smart memory efficient way to store the array
                    # Q = np.fromfile(fid, dt, 3 * V.n1 * V.n2 * V.n3)
                    newpos = (
                        dt.itemsize * nelements + inipos
                    )  # we calculate the offset that we would expect normally with a np.fromfile
                    fid.seek(
                        newpos, os.SEEK_SET
                    )  # we set the file pointer position to this offset

                    V.data[varname + "_X"] = np.transpose(
                        Q[::3].reshape(V.n3, V.n2, V.n1)
                    )
                    V.data[varname + "_Y"] = np.transpose(
                        Q[1::3].reshape(V.n3, V.n2, V.n1)
                    )
                    V.data[varname + "_Z"] = np.transpose(
                        Q[2::3].reshape(V.n3, V.n2, V.n1)
                    )

                else:
                    fid.close()
                    raise ValueError(
                        "In idfxReadVTK: Unknown datatype '%s', should be 'SCALARS' or 'VECTORS'"
                        % datatype
                    )

                logger.debug("read field: {}", varname)

                fid.readline()  # extra line feed
        fid.close()

        return V

    def fargoAdsgReadDat(self, on, *, directory="") -> DataStructure:
        V = DataStructure()
        fluid = "gas"
        densfile = os.path.join(directory, f"{fluid}dens{on}.dat")
        vyfile = os.path.join(directory, f"{fluid}vrad{on}.dat")
        vxfile = os.path.join(directory, f"{fluid}vtheta{on}.dat")

        V.geometry = "polar"
        V.data = {}

        phi = np.loadtxt(os.path.join(directory, "used_azi.dat"))[:, 0]
        domain_x = np.zeros(len(phi) + 1)
        domain_x[:-1] = phi
        domain_x[-1] = 2 * np.pi
        domain_x -= np.pi
        # We avoid ghost cells
        domain_y = np.loadtxt(os.path.join(directory, "used_rad.dat"))
        domain_z = np.zeros(2)

        V.x1 = domain_y  # X-Edge
        V.x2 = domain_x  # Y-Edge
        V.x3 = domain_z  # Z-Edge #latitute

        V.n1 = len(V.x1) - 1  # if len(V.x1)>2 else 2
        V.n2 = len(V.x2) - 1  # if len(V.x2)>2 else 2
        V.n3 = len(V.x3) - 1  # if len(V.x3)>2 else 2
        if Path(densfile).is_file():
            V.data["RHO"] = (
                np.fromfile(densfile, dtype="float64")
                .reshape(V.n3, V.n1, V.n2)
                .transpose(1, 2, 0)
            )  # rad, pĥi, z
        if Path(vyfile).is_file():
            V.data["VX1"] = (
                np.fromfile(vyfile, dtype="float64")
                .reshape(V.n3, V.n1, V.n2)
                .transpose(1, 2, 0)
            )  # rad, pĥi, z
        if Path(vxfile).is_file():
            V.data["VX2"] = (
                np.fromfile(vxfile, dtype="float64")
                .reshape(V.n3, V.n1, V.n2)
                .transpose(1, 2, 0)
            )  # rad, pĥi, z

        return V

    def fargo3dReadDat(
        self, on, *, directory="", inifile="", fluid=None
    ) -> DataStructure:
        if fluid is None:
            fluid = "gas"
        V = DataStructure()
        densfile = os.path.join(directory, f"{fluid}dens{on}.dat")
        vyfile = os.path.join(directory, f"{fluid}vy{on}.dat")
        vxfile = os.path.join(directory, f"{fluid}vx{on}.dat")
        vzfile = os.path.join(directory, f"{fluid}vz{on}.dat")

        if inifile == "":
            params = Parameters(directory=directory, inifile=inifile, code="")
        else:
            params = Parameters(directory=directory, inifile=inifile, code="fargo3d")

        params.loadIniFile()

        class Fargo3DIni:
            def __init__(self, *, COORDINATES, **_kwargs) -> None:
                self.COORDINATES = str(COORDINATES)

        V.geometry = Fargo3DIni(**params.inifile).COORDINATES
        V.data = {}

        domain_x = np.loadtxt(os.path.join(directory, "domain_x.dat"))
        # We avoid ghost cells
        domain_y = np.loadtxt(os.path.join(directory, "domain_y.dat"))[3:-3]
        domain_z = np.loadtxt(os.path.join(directory, "domain_z.dat"))
        if domain_z.shape[0] > 6:
            domain_z = domain_z[3:-3]

        if V.geometry == "cylindrical":
            V.geometry = "polar"
            V.x1 = domain_y  # X-Edge
            V.x2 = domain_x  # Y-Edge
            V.x3 = domain_z  # Z-Edge #latitute

            V.n1 = len(V.x1) - 1  # if len(V.x1)>2 else 2
            V.n2 = len(V.x2) - 1  # if len(V.x2)>2 else 2
            V.n3 = len(V.x3) - 1  # if len(V.x3)>2 else 2
            if Path(densfile).is_file():
                V.data["RHO"] = (
                    np.fromfile(densfile, dtype="float64")
                    .reshape(V.n3, V.n1, V.n2)
                    .transpose(1, 2, 0)
                )  # rad, pĥi, z
            if Path(vyfile).is_file():
                V.data["VX1"] = (
                    np.fromfile(vyfile, dtype="float64")
                    .reshape(V.n3, V.n1, V.n2)
                    .transpose(1, 2, 0)
                )  # rad, pĥi, z
            if Path(vxfile).is_file():
                V.data["VX2"] = (
                    np.fromfile(vxfile, dtype="float64")
                    .reshape(V.n3, V.n1, V.n2)
                    .transpose(1, 2, 0)
                )  # rad, pĥi, z
            if Path(vzfile).is_file():
                V.data["VX3"] = (
                    np.fromfile(vzfile, dtype="float64")
                    .reshape(V.n3, V.n1, V.n2)
                    .transpose(1, 2, 0)
                )  # rad, pĥi, z
            for key in list(V.data.keys()):
                V.data[key] = np.roll(V.data[key], V.n2 // 2, axis=1)
        elif V.geometry == "spherical":
            V.x1 = domain_y  # X-Edge
            V.x2 = domain_z  # Z-Edge #latitute
            V.x3 = domain_x  # Y-Edge

            V.n1 = len(V.x1) - 1  # if len(V.x1)>2 else 2
            V.n2 = len(V.x2) - 1  # if len(V.x2)>2 else 2
            V.n3 = len(V.x3) - 1  # if len(V.x3)>2 else 2
            if Path(densfile).is_file():
                V.data["RHO"] = (
                    np.fromfile(densfile, dtype="float64")
                    .reshape(V.n2, V.n1, V.n3)
                    .transpose(1, 0, 2)
                )  # rad, pĥi, z
            if Path(vyfile).is_file():
                V.data["VX1"] = (
                    np.fromfile(vyfile, dtype="float64")
                    .reshape(V.n2, V.n1, V.n3)
                    .transpose(1, 0, 2)
                )  # rad, pĥi, z
            if Path(vzfile).is_file():
                V.data["VX2"] = (
                    np.fromfile(vzfile, dtype="float64")
                    .reshape(V.n2, V.n1, V.n3)
                    .transpose(1, 0, 2)
                )  # rad, pĥi, z
            if Path(vxfile).is_file():
                V.data["VX3"] = (
                    np.fromfile(vxfile, dtype="float64")
                    .reshape(V.n2, V.n1, V.n3)
                    .transpose(1, 0, 2)
                )  # rad, pĥi, z
            for key in list(V.data.keys()):
                V.data[key] = np.roll(V.data[key], V.n2 // 2, axis=2)
        else:
            raise ValueError(f"{V.geometry} not implemented yet for fargo3d.")

        if not V.data:
            raise FileNotFoundError(f"No file matches the pattern '{fluid}*{on}.dat'")

        return V



__all__ = [
    "PathT",
    "StrDict",
    "FloatArray",
    "Geometry",
    "FrameType",
    "BinData",
    "OrbitalElements",
    "PlanetData",
    "IniData",
    "BinReader",
    "PlanetReader",
    "IniReader",
]
import sys
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import Any, Protocol, Union, final

import numpy as np

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

if sys.version_info >= (3, 11):
    from enum import StrEnum
    from typing import assert_never
else:
    from typing_extensions import assert_never

    from nonos._backports import StrEnum


PathT: TypeAlias = Union[str, Path]
StrDict: TypeAlias = dict[str, Any]

FloatArray: TypeAlias = "np.ndarray[Any, np.dtype[np.float32 | np.float64]]"


class Geometry(StrEnum):
    CARTESIAN = auto()
    POLAR = auto()
    SPHERICAL = auto()


class FrameType(Enum):
    FIXED_FRAME = auto()
    CONSTANT_ROTATION = auto()
    PLANET_COROTATION = auto()


@final
@dataclass(frozen=True, eq=False)
class BinData:
    # TODO: use slots=True in @dataclass when Python 3.9 is dropped
    __slots__ = ["data", "geometry", "x1", "x2", "x3"]
    data: StrDict
    geometry: Geometry
    x1: FloatArray
    x2: FloatArray
    x3: FloatArray

    @classmethod
    def default_init(cls):
        return {
            field.name: field.default for field in cls.__dataclass_fields__.values()
        }


@final
@dataclass(frozen=True, eq=False)
class OrbitalElements:
    # TODO: use slots=True in @dataclass when Python 3.9 is dropped
    __slots__ = ["i", "e", "a"]
    i: FloatArray
    e: FloatArray
    a: FloatArray


@final
@dataclass(frozen=True, eq=False)
class PlanetData:
    # fields that are required at __init__
    _init_attrs = ["x", "y", "z", "vx", "vy", "vz", "q", "t", "dt"]
    # additional derived field that can be computed on the fly
    _post_init_attrs = ["d"]
    __slots__ = _init_attrs + _post_init_attrs

    # cartesian position
    x: FloatArray
    y: FloatArray
    z: FloatArray

    # cartesian velocity
    vx: FloatArray
    vy: FloatArray
    vz: FloatArray

    # mass ratio (or mass in units of the central star's)
    q: FloatArray

    # time and timestep
    t: FloatArray
    dt: FloatArray

    def __post_init__(self) -> None:
        object.__setattr__(self, "d", np.sqrt(self.x**2 + self.y**2 + self.z**2))

    def get_orbital_elements(self, frame: FrameType) -> OrbitalElements:
        if frame is FrameType.FIXED_FRAME:
            hx = self.y * self.vz - self.z * self.vy
            hy = self.z * self.vx - self.x * self.vz
            hz = self.x * self.vy - self.y * self.vx
            hhor = np.hypot(hx, hy)

            h2 = hx * hx + hy * hy + hz * hz
            h = np.sqrt(h2)
            i = np.arcsin(hhor / h)

            d = object.__getattribute__(self, "d")
            Ax = self.vy * hz - self.vz * hy - (1.0 + self.q) * self.x / d
            Ay = self.vz * hx - self.vx * hz - (1.0 + self.q) * self.y / d
            Az = self.vx * hy - self.vy * hx - (1.0 + self.q) * self.z / d

            e = np.sqrt(Ax * Ax + Ay * Ay + Az * Az) / (1.0 + self.q)
            a = h * h / ((1.0 + self.q) * (1.0 - e * e))
            return OrbitalElements(i, e, a)
        elif frame is FrameType.CONSTANT_ROTATION:
            raise NotImplementedError(
                f"PlanetData.set_orbital_elements isn't implemented for {frame=}"
            )
        elif frame is FrameType.PLANET_COROTATION:
            # bug-for-bug compat
            return self.get_orbital_elements(FrameType.FIXED_FRAME)
        else:
            assert_never(frame)

    def get_rotational_rate(self) -> FloatArray:
        d = self.d  # type: ignore [attr-defined]
        return np.sqrt((1.0 + self.q) / pow(d, 3.0))


for key in PlanetData._post_init_attrs:
    PlanetData.__annotations__[key] = FloatArray


@final
@dataclass(frozen=True)
class IniData:
    # TODO: use slots=True in @dataclass when Python 3.9 is dropped
    __slots__ = [
        "file",
        "frame",
        "rotational_rate",
        "output_time_interval",
        "meta",
    ]
    file: Path
    frame: FrameType
    rotational_rate: float
    output_time_interval: float
    meta: StrDict


class BinReader(Protocol):
    @staticmethod
    def parse_output_number_and_filename(
        file_or_number: Union[PathT, int],
        *,
        directory: PathT,
        prefix: str,
    ) -> tuple[int, Path]: ...

    @staticmethod
    def get_bin_files(directory: PathT, /) -> list[Path]: ...

    @staticmethod
    def read(file: PathT, /, **meta) -> BinData: ...


class PlanetReader(Protocol):
    @staticmethod
    def get_planet_files(directory: Path, /) -> list[Path]: ...

    @staticmethod
    def read(file: PathT, /) -> PlanetData: ...


class IniReader(Protocol):
    @staticmethod
    def read(file: PathT, /) -> IniData: ...



__all__ = [
    "IdefixVTKReader",
    "PlutoVTKReader",
    "Fargo3DReader",
    "FargoADSGReader",
]
from pathlib import Path
from typing import final

import inifix

from nonos._readers._base import ReaderMixin
from nonos._types import FrameType, IniData, PathT


@final
class IdefixVTKReader(ReaderMixin):
    @staticmethod
    def read(file: PathT, /) -> IniData:
        class IdefixIniOutput:
            def __init__(self, *, vtk, **_kwargs) -> None:
                self.vtk = float(vtk)

        class IdefixIniHydro:
            def __init__(self, **kwargs) -> None:
                if "rotation" in kwargs:
                    self.frame = FrameType.CONSTANT_ROTATION
                    self.rotation = float(kwargs["rotation"])
                else:
                    self.frame = FrameType.FIXED_FRAME
                    self.rotation = 0.0

        class IdefixIni:
            def __init__(self, *, Hydro, Output, **_kwargs) -> None:
                self.hydro = IdefixIniHydro(**Hydro)
                self.output = IdefixIniOutput(**Output)

        meta = inifix.load(file)
        ini = IdefixIni(**meta)

        return IniData(
            file=Path(file).resolve(),
            frame=ini.hydro.frame,
            rotational_rate=ini.hydro.rotation,
            output_time_interval=ini.output.vtk,
            meta=meta,
        )


@final
class PlutoVTKReader(ReaderMixin):
    @staticmethod
    def read(file: PathT, /) -> IniData:
        class PlutoIniOutput:
            def __init__(self, *, vtk, **_kwargs) -> None:
                self.vtk = float(list(vtk)[0])

        class PlutoIni:
            def __init__(self, **kwargs) -> None:
                self.output = PlutoIniOutput(**kwargs["Static Grid Output"])

        meta = inifix.load(file)
        ini = PlutoIni(**meta)

        return IniData(
            file=Path(file).resolve(),
            frame=FrameType.FIXED_FRAME,
            rotational_rate=0.0,
            output_time_interval=ini.output.vtk,
            meta=meta,
        )


@final
class Fargo3DReader(ReaderMixin):
    @staticmethod
    def read(file: PathT, /) -> IniData:
        class Fargo3DIni:
            def __init__(
                self,
                *,
                NINTERM,
                DT,
                FRAME: str = "F",
                OMEGAFRAME: float = 0.0,
                **_kwargs,
            ) -> None:
                self.NINTERM = int(NINTERM)
                self.DT = float(DT)
                self.FRAME: FrameType
                self.OMEGAFRAME: float

                if FRAME == "F":
                    self.OMEGAFRAME = float(OMEGAFRAME)
                    if self.OMEGAFRAME == 0.0:
                        self.FRAME = FrameType.FIXED_FRAME
                    else:
                        self.FRAME = FrameType.CONSTANT_ROTATION
                elif FRAME == "C":
                    self.FRAME = FrameType.PLANET_COROTATION
                    self.OMEGAFRAME = float("nan")
                else:
                    raise NotImplementedError

        meta = inifix.load(file)
        ini = Fargo3DIni(**meta)

        return IniData(
            file=Path(file).resolve(),
            frame=ini.FRAME,
            rotational_rate=ini.OMEGAFRAME,
            output_time_interval=ini.NINTERM * ini.DT,
            meta=meta,
        )


@final
class FargoADSGReader(ReaderMixin):
    @staticmethod
    def read(file: PathT, /) -> IniData:
        class FargoADSGIni:
            def __init__(
                self,
                *,
                Ninterm,
                DT,
                Frame: str = "F",
                OmegaFrame: float = 0.0,
                **_kwargs,
            ) -> None:
                self.NINTERM = int(Ninterm)
                self.DT = float(DT)
                self.FRAME: FrameType
                self.OMEGAFRAME: float

                if Frame == "F":
                    self.OMEGAFRAME = float(OmegaFrame)
                    if self.OMEGAFRAME == 0.0:
                        self.FRAME = FrameType.FIXED_FRAME
                    else:
                        self.FRAME = FrameType.CONSTANT_ROTATION
                elif Frame == "C":
                    self.FRAME = FrameType.PLANET_COROTATION
                    self.OMEGAFRAME = float("nan")
                else:
                    raise NotImplementedError

        meta = inifix.load(file)
        ini = FargoADSGIni(**meta)

        return IniData(
            file=Path(file).resolve(),
            frame=ini.FRAME,
            rotational_rate=ini.OMEGAFRAME,
            output_time_interval=ini.NINTERM * ini.DT,
            meta=meta,
        )



import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Optional

import numpy as np
from lick.lick import lick_box

from nonos.api.analysis import Coordinates, GasField, Plotable
from nonos.loaders import Recipe, loader_from, recipe_from

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure

    from nonos._types import FloatArray, PathT


def file_analysis(
    filename: "PathT",
    *,
    inifile: Optional[str] = None,
    code: Optional[str] = None,
    directory: Optional["PathT"] = None,
    norb: Optional[int] = None,
) -> "FloatArray":
    from scipy.ndimage import uniform_filter1d

    if directory is None:
        directory = Path.cwd()
    else:
        directory = Path(directory)

    columns = np.loadtxt(directory / filename, dtype="float64").T
    if norb is None:
        return columns

    loader = loader_from(
        code=code,
        parameter_file=inifile,
        directory=directory,
    )
    recipe = recipe_from(code=code, parameter_file=inifile, directory=directory)
    ini = loader.load_ini_file().meta

    if recipe is Recipe.IDEFIX_VTK and "analysis" in ini["Output"]:
        analysis = ini["Output"]["analysis"]
        rpini = ini["Planet"]["dpl"]
        Ntmean = round(norb * 2 * np.pi * pow(rpini, 1.5) / analysis)
        for i in range(1, len(columns) - 1):
            columns[i] = uniform_filter1d(columns[i], Ntmean)
    else:
        raise NotImplementedError(
            f"moving average on {norb} orbits is not implemented for the recipe {recipe}"
        )
    return columns


def planet_analysis(
    planet_number, *, inifile="", code="", directory="", norb=None
):  # pragma: no cover
    from nonos.api.from_simulation import Parameters

    warnings.warn(
        "nonos.api.satellite.planet_analysis is deprecated and will be removed in "
        "a future version. Please use nonos.api.satellite.file_analysis instead.",
        category=DeprecationWarning,
        stacklevel=2,
    )
    if norb is not None:
        warnings.warn(
            "The norb argument has no effect and is deprecated",
            stacklevel=2,
        )
    init = Parameters(inifile=inifile, code=code, directory=directory)
    init.loadIniFile()
    init.loadPlanetFile(planet_number=planet_number)
    return init


class NonosLick:
    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
        lx: GasField,
        ly: GasField,
        field: GasField,
        *,
        xmin: Optional[float] = None,
        xmax: Optional[float] = None,
        ymin: Optional[float] = None,
        ymax: Optional[float] = None,
        size_interpolated: int = 1000,
        niter_lic: int = 6,
        kernel_length: int = 101,
        method: str = "linear",
        method_background: str = "nearest",
        light_source: bool = True,
    ):
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax

        # (x,y) are 2D meshgrids at cell centers
        self.X, self.Y, self.LINE1, self.LINE2, self.F, self.lick = lick_box(
            x,
            y,
            lx.data,
            ly.data,
            field.data,
            xmin=self.xmin,
            xmax=self.xmax,
            ymin=self.ymin,
            ymax=self.ymax,
            size_interpolated=size_interpolated,
            niter_lic=niter_lic,
            kernel_length=kernel_length,
            method=method,
            method_background=method_background,
            light_source=light_source,
        )

    def plot(
        self,
        fig: "Figure",
        ax: "Axes",
        *,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        alpha: float = 0.45,
        log: bool = False,
        cmap=None,
        title: Optional[str] = None,
        density_streamlines: Optional[float] = None,
        color_streamlines: str = "black",
    ):
        dict_background = {}
        dict_background["field"] = "background"
        dict_background["abscissa"] = "x"
        dict_background["ordinate"] = "y"
        dict_background[dict_background["field"]] = self.F
        dict_background[dict_background["abscissa"]] = self.X
        dict_background[dict_background["ordinate"]] = self.Y

        dict_lick = {}
        dict_lick["field"] = "lick"
        dict_lick["abscissa"] = "x"
        dict_lick["ordinate"] = "y"
        dict_lick[dict_lick["field"]] = self.lick
        dict_lick[dict_lick["abscissa"]] = self.X
        dict_lick[dict_lick["ordinate"]] = self.Y

        im = Plotable(dict_background).plot(
            fig,
            ax,
            vmin=vmin,
            vmax=vmax,
            log=log,
            cmap=cmap,
            filename=None,
            dpi=500,
            title=title,
            shading="nearest",
            rasterized=True,
        )
        Plotable(dict_lick).plot(
            fig,
            ax,
            log=False,
            cmap="binary_r",
            filename=None,
            dpi=500,
            title=None,
            alpha=alpha,
            shading="nearest",
            rasterized=True,
        )
        if density_streamlines is not None:
            ax.streamplot(
                self.X,
                self.Y,
                self.LINE1,
                self.LINE2,
                density=density_streamlines,
                arrowstyle="->",
                linewidth=0.5,
                color=color_streamlines,
                # color=np.log10(self.F*np.sqrt(self.LINE1**2+self.LINE2**2)),#/np.max(np.log10(self.F*np.sqrt(self.LINE1**2+self.LINE2**2))),
                # cmap=cb.cbmap("binary_r"),
            )
        ax.set_xlim(self.xmin, self.xmax)
        ax.set_ylim(self.ymin, self.ymax)
        return im


def compute(
    field: str,
    data: np.ndarray,
    ref: GasField,
):
    ret_data = data
    ret_coords = ref.coords
    geometry = ret_coords.geometry
    return GasField(
        field,
        ret_data,
        ret_coords,
        geometry,
        ref.on,
        operation=ref.operation,
        inifile=ref.inifile,
        directory=ref.directory,
        rotate_by=ref._rotate_by,
    )


def from_data(
    *,
    field: str,
    data: np.ndarray,
    coords: Coordinates,
    on: int,
    operation: str,
    inifile: Optional[str] = None,
    code: Optional[str] = None,
    directory: str = "",
    rotate_grid: int = -1,
):  # pragma: no cover
    warnings.warn(
        "nonos.api.satellite.from_data is deprecated and will be removed "
        "in a future version. Please use nonos.api.satellite.compute instead",
        category=DeprecationWarning,
        stacklevel=2,
    )
    ret_data = data
    ret_coords = coords
    geometry = coords.geometry
    return GasField(
        field,
        ret_data,
        ret_coords,
        geometry,
        on,
        operation=operation,
        inifile=inifile,
        code=code,
        directory=directory,
        rotate_grid=rotate_grid,
    )


def from_file(
    *,
    field: str,
    operation: str,
    on: int,
    directory: Optional["PathT"] = None,
):
    if directory is None:
        directory = Path.cwd()
    else:
        directory = Path(directory)
    repout = field.lower()
    headername = directory / "header" / f"header_{operation}.npy"
    with open(headername, "rb") as file:
        dict_coords = np.load(file, allow_pickle=True).item()

    geometry, coord0, coord1, coord2 = dict_coords.values()
    ret_coords = Coordinates(geometry, coord0, coord1, coord2)

    fileout = directory / repout / f"_{operation}_{field}.{on:04d}.npy"
    with open(fileout, "rb") as file:
        ret_data = np.load(file, allow_pickle=True)

    return GasField(
        field,
        ret_data,
        ret_coords,
        geometry,
        on,
        operation=operation,
        directory=directory,
    )



import importlib.resources as importlib_resources
from pathlib import Path
from typing import cast

import matplotlib as mpl


def scale_mpl(scaling: float) -> None:
    # Scale all the parameters by the same factor depending on the level
    # this heavily borrows from seaborn.set_context (v0.11.1) see
    # https://github.com/mwaskom/seaborn/blob/a41703e7fddf8f66b1fd5f994f983b37e865a3b2/seaborn/rcmod.py#L439
    # https://github.com/mwaskom/seaborn/blob/a41703e7fddf8f66b1fd5f994f983b37e865a3b2/seaborn/rcmod.py#L338

    # Set up dictionary of default parameters
    texts_base_context = {
        "font.size": 12,
        "axes.labelsize": 12,
        "axes.titlesize": 12,
        "xtick.labelsize": 11,
        "ytick.labelsize": 11,
        "legend.fontsize": 11,
        "legend.title_fontsize": 12,
    }

    base_context = {
        "axes.linewidth": 1.25,
        "grid.linewidth": 1,
        "lines.linewidth": 1.5,
        "lines.markersize": 6,
        "patch.linewidth": 1,
        "xtick.major.width": 1.25,
        "ytick.major.width": 1.25,
        "xtick.minor.width": 1,
        "ytick.minor.width": 1,
        "xtick.major.size": 6,
        "ytick.major.size": 6,
        "xtick.minor.size": 4,
        "ytick.minor.size": 4,
    }
    base_context.update(texts_base_context)

    context_dict = {k: v * scaling for k, v in base_context.items()}

    # the reason why the scaling are separated comes
    # from seaborn where the font sizes are controled
    # by an independent factor, so I'm keeping the structure
    # in case we want to do that as well later
    font_dict = {k: context_dict[k] * scaling for k in texts_base_context}
    context_dict.update(font_dict)
    mpl.rcParams.update(context_dict)


def set_mpl_style(scaling: float) -> None:
    if mpl.__version_info__ >= (3, 7):
        import matplotlib.pyplot as plt

        plt.style.use("nonos.default")
    else:
        # promise mypy this is a Path to get around a broad return type
        # from importlib_resource.files
        nonos_path = cast(Path, importlib_resources.files("nonos"))
        mpl.rc_file(nonos_path / "default.mplstyle")
    scale_mpl(scaling)



import re
from itertools import combinations
from typing import Optional

import numpy as np
import pytest

from nonos.api._angle_parsing import _parse_planet_file, _parse_rotation_angle
from nonos.parsing import (
    parse_image_format,
    parse_output_number_range,
    parse_range,
    range_converter,
    userval_or_default,
)


@pytest.mark.parametrize(
    "received, expected",
    [
        (("val", "default"), "val"),
        (("unset", "default"), "default"),
        ((None, "default"), "default"),
        (("unset", None), None),
        ((None, None), None),
    ],
)
def test_userval_or_default(received, expected):
    userval, default = received
    assert userval_or_default(userval, default=default) == expected


@pytest.mark.parametrize(
    "received, expected",
    [
        (1, [1]),
        (0, [0]),
        ([1], [1]),
        ([0, 1], [0, 1]),
        ([0, 2], [0, 1, 2]),
        ([0, 5], [0, 1, 2, 3, 4, 5]),
        ([10, 20, 5], [10, 15, 20]),
        ([0, 299, 50], [0, 50, 100, 150, 200, 250]),
    ],
)
def test_range_outputs(received, expected):
    assert parse_output_number_range(received) == expected


def test_from_maxval():
    maxval = 1.0
    assert parse_output_number_range(None, maxval=maxval) == [maxval]


def test_unparseable_data():
    with pytest.raises(
        ValueError, match="Can't parse a range from unset values without a max."
    ):
        parse_output_number_range("unset")


def test_invalid_request():
    with pytest.raises(
        ValueError, match="No output beyond 5 is available, but 10 was requested."
    ):
        parse_output_number_range([1, 10], maxval=5)


def test_invalid_nargs():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Can't parse a range from sequence [1, 2, 3, 4] with more than 3 values."
        ),
    ):
        parse_output_number_range([1, 2, 3, 4])


@pytest.mark.parametrize(
    "dim, extent",
    [
        (2, ("0.4", "8", "0", "-0.2", "0.2")),
        (1, (0.4, 8, 0.2)),
    ],
)
def test_invalid_nargs_parse_range(dim, extent):
    with pytest.raises(
        ValueError,
        match=re.escape(
            f"Received sequence `extent` with incorrect size {len(extent)}. Expected exactly {2*dim=} values."
        ),
    ):
        parse_range(extent, dim)


@pytest.mark.parametrize(
    "received",
    [
        [1, 0],
        [1, 0, 1],
        [1, 0, 2],
    ],
)
def test_invalid_range(received):
    with pytest.raises(ValueError, match="Can't parse a range with max < min."):
        parse_output_number_range(received)


@pytest.mark.parametrize(
    "abscissa, ordinate, dim, expected",
    [
        (
            np.linspace(0.2, 10, 100),
            np.linspace(-np.pi, np.pi, 100),
            2,
            (0.2, 10.0, -np.pi, np.pi),
        ),
        (np.linspace(0.2, 10, 100), np.zeros(2), 1, (0.2, 10.0)),
    ],
)
def test_parse_range(abscissa, ordinate, dim, expected):
    extent1 = parse_range("unset", dim=dim)
    assert range_converter(extent1, abscissa=abscissa, ordinate=ordinate) == expected
    extent2 = parse_range(("0.5", "5", "-0.2", "0.2"), dim=2)
    assert range_converter(
        extent2,
        abscissa=np.linspace(0.2, 10, 100),
        ordinate=np.linspace(-0.4, 0.4, 100),
    ) == (0.5, 5, -0.2, 0.2)
    extent3 = parse_range(("0.4", "9.5"), dim=1)
    assert range_converter(
        extent3, abscissa=np.linspace(0.2, 10, 100), ordinate=np.zeros(2)
    ) == (0.4, 9.5)
    extent4 = parse_range(("0.5", "x", "-0.2", "x"), dim=2)
    assert extent4 == (0.5, None, -0.2, None)
    assert range_converter(
        extent4,
        abscissa=np.linspace(0.2, 10, 100),
        ordinate=np.linspace(-0.4, 0.4, 100),
    ) == (0.5, 10.0, -0.2, 0.4)


def test_range_converter_error():
    with pytest.raises(
        TypeError,
        match="Expected extent to be of lenght 2 or 4,",
    ):
        range_converter(extent=[], abscissa=None, ordinate=None)


@pytest.mark.parametrize(
    "received, expected",
    [
        (".png", "png"),
        (".pdf", "pdf"),
        ("png", "png"),
        ("pdf", "pdf"),
    ],
)
def test_image_format(received, expected):
    assert parse_image_format(received) == expected


@pytest.mark.parametrize("received", ["unset", None])
def test_image_format_default(received):
    result = parse_image_format(received)
    assert isinstance(result, str)
    assert re.fullmatch(r"\w+", result)


def test_invalid_image_format():
    fake_ext = ".pnd"
    with pytest.raises(
        ValueError,
        match=f"^(Received unknown file format '{fake_ext}'. Available formated are)",
    ):
        parse_image_format(fake_ext)


class TestParsePlanetFile:
    def test_from_filename(self):
        input_ = "test"
        assert _parse_planet_file(planet_file=input_) == input_

    def test_from_number(self):
        input_ = 456
        assert _parse_planet_file(planet_number=input_) == f"planet{input_}.dat"

    def test_both_args(self):
        with pytest.raises(TypeError):
            _parse_planet_file(
                planet_file="test",
                planet_number=1,
            )

    def test_no_args(self):
        assert _parse_planet_file() == "planet0.dat"


def mock_find_phip(
    planet_number: Optional[int] = None,  # noqa: ARG001
    *,
    planet_file: Optional[str] = None,  # noqa: ARG001
) -> float:
    return 0.0


class TestParseRotationAngle:
    example_inputs = {
        "rotate_by": 1.0,
        "rotate_with": "planet0.dat",
        "planet_number_argument": ("test", 0),
    }
    default_kwargs = {
        "rotate_by": None,
        "rotate_with": None,
        "planet_number_argument": ("test", None),
        "find_phip": mock_find_phip,
        "stacklevel": 2,
    }

    @pytest.mark.parametrize("kwargs", combinations(example_inputs.items(), 2))
    def test_two_inputs(self, kwargs):
        conf = {**self.default_kwargs, **dict(kwargs)}
        with pytest.raises(TypeError, match="Can only process one argument"):
            _parse_rotation_angle(**conf)

    def test_all_inputs(self):
        conf = {**self.default_kwargs, **self.example_inputs}
        with pytest.raises(TypeError, match="Can only process one argument"):
            _parse_rotation_angle(**conf)

    def test_from_rotate_with(self):
        conf = {
            **self.default_kwargs,
            "rotate_with": self.example_inputs["rotate_with"],
        }
        result = _parse_rotation_angle(**conf)
        assert result == 0.0

    def test_from_planet_number(self):
        conf = {
            **self.default_kwargs,
            "planet_number_argument": self.example_inputs["planet_number_argument"],
        }
        with pytest.deprecated_call():
            result = _parse_rotation_angle(**conf)
        assert result == 0.0



__all__ = [
    "VTKReader",
    "Fargo3DReader",
    "FargoADSGReader",
]
import json
import os
import re
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, Union, final

import numpy as np

from nonos._readers._base import ReaderMixin
from nonos._types import BinData, FloatArray, Geometry, PathT

if sys.version_info >= (3, 11):
    from typing import assert_never
else:
    from typing_extensions import assert_never

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


@final
class VTKReader(ReaderMixin):
    @staticmethod
    def parse_output_number_and_filename(
        file_or_number: Union[PathT, int],
        *,
        directory: PathT,
        prefix: str,  # noqa: ARG004
    ) -> tuple[int, Path]:
        if isinstance(file_or_number, (str, Path)):
            file = Path(file_or_number)
            if (m := re.search(r"\d+", file.name)) is None:
                raise ValueError(
                    f"Failed to parse output number from filename {file.name}"
                )
            else:
                output_number = int(m.group())
        else:
            output_number = int(file_or_number)
            file = Path(f"data.{output_number:04d}.vtk")

        if file == Path(file.name):
            file = Path(directory) / file

        return output_number, file

    @staticmethod
    def get_bin_files(directory: PathT, /) -> list[Path]:
        directory = Path(directory)
        return sorted(directory.glob("data.*.vtk"))

    @staticmethod
    def read(file, /, **meta) -> BinData:
        """
        Adapted from Geoffroy Lesur
        Function that reads a vtk file in polar coordinates
        """

        meta.setdefault("geometry", None)
        meta.setdefault("cell", "edges")
        meta.setdefault("computedata", True)

        fid = open(file, "rb")

        # define our datastructure
        V = BinData.default_init()

        # raw data which will be read from the file
        V["data"] = {}

        # initialize geometry
        if meta["geometry"] is not None:
            V["geometry"] = Geometry(meta["geometry"])

        # datatype we read
        dt = np.dtype(">f")  # Big endian single precision floats
        dint = np.dtype(">i4")  # Big endian integer

        s = fid.readline()  # VTK DataFile Version x.x
        s = fid.readline()  # Comments

        s = fid.readline()  # BINARY
        s = fid.readline()  # DATASET RECTILINEAR_GRID or STRUCTURED_GRID
        slist = s.split()

        s = fid.readline()  # DIMENSIONS NX NY NZ
        slist = s.split()
        entry = str(slist[0], "utf-8")
        if entry == "FIELD":
            nfield = int(slist[2])
            for _field in range(nfield):
                s = fid.readline()
                slist = s.split()
                entry = str(slist[0], "utf-8")
                if entry == "TIME":
                    # skip
                    fid.seek(dt.itemsize, os.SEEK_CUR)
                elif entry == "GEOMETRY":
                    g = np.fromfile(fid, dint, 1)
                    if g == 0:
                        thisgeometry = Geometry.CARTESIAN
                    elif g == 1:
                        thisgeometry = Geometry.POLAR
                    elif g == 2:
                        thisgeometry = Geometry.SPHERICAL
                    elif g == 3:  # pragma: co cover
                        fid.close()
                        raise NotImplementedError(
                            "Support for cylindrical geometry is missing"
                        )
                    else:  # pragma: no cover
                        fid.close()
                        raise ValueError(
                            f"Unknown value for GEOMETRY flag ({g!r}) was found in the VTK file."
                        )

                    if meta["geometry"] is not None:
                        # We already have a proposed geometry, check that what is read from the file matches
                        if thisgeometry != meta["geometry"]:  # pragma: no cover
                            fid.close()
                            raise ValueError(
                                f"geometry argument ({meta['geometry']!r}) is "
                                "inconsistent with GEOMETRY flag from the VTK file "
                                f"({thisgeometry!r})"
                            )
                    V["geometry"] = thisgeometry
                elif entry == "PERIODICITY":
                    # skip
                    fid.seek(dint.itemsize * 3, os.SEEK_CUR)
                else:  # pragma: no cover
                    fid.close()
                    raise ValueError(f"Received unknown field: {entry!r}")

                s = fid.readline()  # extra linefeed

            # finished reading the field entry
            # read next line
            s = fid.readline()  # DIMENSIONS...

        if not isinstance(V["geometry"], Geometry):  # pragma: no cover
            fid.close()
            raise RuntimeError(
                "Geometry couldn't be determined from data. "
                "Try to set the geometry keyword argument explicitly."
            )
        slist = s.split()  # DIMENSIONS....
        n1 = int(slist[1])
        n2 = int(slist[2])
        n3 = int(slist[3])

        z: Union[np.ndarray, np.memmap]

        if V["geometry"] is Geometry.CARTESIAN:
            s = fid.readline()  # X_COORDINATES NX float
            # we store the file pointer position before computing points
            inipos = fid.tell()
            # some smart memory efficient way to store the array
            x = np.memmap(fid, mode="r", dtype=dt, offset=inipos, shape=n1)
            # we calculate the offset that we would expect normally with a np.fromfile
            newpos = dt.itemsize * n1 + inipos
            # we set the file pointer position to this offset
            fid.seek(newpos, os.SEEK_SET)
            s = fid.readline()  # Extra line feed added by idefix

            s = fid.readline()  # Y_COORDINATES NY float
            # we store the file pointer position before computing points
            inipos = fid.tell()
            # some smart memory efficient way to store the array
            y = np.memmap(fid, mode="r", dtype=dt, offset=inipos, shape=n2)
            # we calculate the offset that we would expect normally with a np.fromfile
            newpos = dt.itemsize * n2 + inipos
            # we set the file pointer position to this offset
            fid.seek(newpos, os.SEEK_SET)
            s = fid.readline()  # Extra line feed added by idefix

            s = fid.readline()  # Z_COORDINATES NZ float
            # we store the file pointer position before computing points
            inipos = fid.tell()
            # some smart memory efficient way to store the array
            z = np.memmap(fid, mode="r", dtype=dt, offset=inipos, shape=n3)
            # we calculate the offset that we would expect normally with a np.fromfile
            newpos = dt.itemsize * 1 * n3 + inipos
            # we set the file pointer position to this offset
            fid.seek(newpos, os.SEEK_SET)
            s = fid.readline()  # Extra line feed added by idefix

            s = fid.readline()  # POINT_DATA NXNYNZ

            slist = s.split()
            point_type = str(slist[0], "utf-8")
            npoints = int(slist[1])
            s = fid.readline()  # EXTRA LINE FEED

            if point_type == "CELL_DATA" or meta["cell"] == "centers":
                # The file contains face coordinates, so we extrapolate to get the cell center coordinates.
                if n1 > 1:
                    n1 -= 1
                    if meta["cell"] == "centers":
                        V["x1"] = 0.5 * (x[1:] + x[:-1])
                    elif meta["cell"] == "edges":
                        V["x1"] = x
                else:
                    V["x1"] = x
                if n2 > 1:
                    n2 -= 1
                    if meta["cell"] == "centers":
                        V["x2"] = 0.5 * (y[1:] + y[:-1])
                    elif meta["cell"] == "edges":
                        V["x2"] = y
                else:
                    V["x2"] = y
                if n3 > 1:
                    n3 -= 1
                    if meta["cell"] == "centers":
                        V["x3"] = 0.5 * (z[1:] + z[:-1])
                    elif meta["cell"] == "edges":
                        V["x3"] = z
                else:
                    V["x3"] = z
            elif point_type == "POINT_DATA":
                V["x1"] = x
                V["x2"] = y
                V["x3"] = z

            if (grid_size := n1 * n2 * n3) != npoints:  # pragma: no cover
                fid.close()
                raise ValueError(
                    f"Grid size ({grid_size}) is not consistent with number "
                    f"of points ({npoints}) in the data set"
                )
            del grid_size

        elif V["geometry"] is Geometry.POLAR or V["geometry"] is Geometry.SPHERICAL:
            if n3 == 1:
                is2d = 1
            else:
                is2d = 0

            s = fid.readline()  # POINTS NXNYNZ float
            slist = s.split()
            npoints = int(slist[1])

            # we store the file pointer position before computing points
            inipos = fid.tell()

            # some smart memory efficient way to store the array
            points = np.memmap(
                fid, mode="r", dtype=dt, offset=inipos, shape=3 * npoints
            )
            # we calculate the offset that we would expect normally with a np.fromfile
            newpos = dt.itemsize * 3 * npoints + inipos
            # we set the file pointer position to this offset
            fid.seek(newpos, os.SEEK_SET)
            s = fid.readline()  # EXTRA LINE FEED

            if (grid_size := n1 * n2 * n3) != npoints:  # pragma: no cover
                fid.close()
                raise ValueError(
                    f"Grid size ({grid_size}) is not consistent with number "
                    f"of points ({npoints}) in the data set"
                )
            del grid_size

            # Reconstruct the polar coordinate system
            x1d = points[::3]
            y1d = points[1::3]
            z1d = points[2::3]

            new_shape = n3, n2, n1
            xcart = np.transpose(x1d.reshape(new_shape))
            ycart = np.transpose(y1d.reshape(new_shape))
            zcart = np.transpose(z1d.reshape(new_shape))
            del new_shape

            # Reconstruct the polar coordinate system
            if V["geometry"] is Geometry.POLAR:
                r = np.sqrt(xcart[:, 0, 0] ** 2 + ycart[:, 0, 0] ** 2)
                theta = np.unwrap(np.arctan2(ycart[0, :, 0], xcart[0, :, 0]))
                z = zcart[0, 0, :]

                s = fid.readline()  # CELL_DATA (NX-1)(NY-1)(NZ-1)
                slist = s.split()
                data_type = str(slist[0], "utf-8")
                if data_type != "CELL_DATA":  # pragma: no cover
                    fid.close()
                    raise ValueError(
                        f"Expected 'CELL DATA' as produced by PLUTO, got {data_type!r}."
                    )
                s = fid.readline()  # Line feed

                if meta["cell"] == "edges":
                    if n1 > 1:
                        n1 -= 1
                        V["x1"] = r
                    else:
                        V["x1"] = r
                    if n2 > 1:
                        n2 -= 1
                        V["x2"] = theta
                    else:
                        V["x2"] = theta
                    if n3 > 1:
                        n3 -= 1
                        V["x3"] = z
                    else:
                        V["x3"] = z

                # Perform averaging on coordinate system to get cell centers
                # The file contains face coordinates, so we extrapolate to get the cell center coordinates.
                elif meta["cell"] == "centers":
                    if n1 > 1:
                        n1 -= 1
                        V["x1"] = 0.5 * (r[1:] + r[:-1])
                    else:
                        V["x1"] = r
                    if n2 > 1:
                        n2 -= 1
                        V["x2"] = (0.5 * (theta[1:] + theta[:-1]) + np.pi) % (
                            2.0 * np.pi
                        ) - np.pi
                    else:
                        V["x2"] = theta
                    if n3 > 1:
                        n3 -= 1
                        V["x3"] = 0.5 * (z[1:] + z[:-1])
                    else:
                        V["x3"] = z

            # Reconstruct the spherical coordinate system
            elif V["geometry"] is Geometry.SPHERICAL:
                if is2d:
                    r = np.sqrt(xcart[:, 0, 0] ** 2 + zcart[:, 0, 0] ** 2)
                    phi = np.unwrap(
                        np.arctan2(ycart[0, n2 // 2, :], xcart[0, n2 // 2, :])
                    )
                    theta = np.arccos(
                        zcart[0, :, 0]
                        / np.sqrt(xcart[0, :, 0] ** 2 + zcart[0, :, 0] ** 2)
                    )
                else:
                    r = np.sqrt(
                        xcart[:, 0, 0] ** 2 + ycart[:, 0, 0] ** 2 + zcart[:, 0, 0] ** 2
                    )
                    phi = np.unwrap(
                        np.arctan2(
                            ycart[n1 // 2, n2 // 2, :],
                            xcart[n1 // 2, n2 // 2, :],
                        )
                    )
                    theta = np.arccos(
                        zcart[0, :, 0]
                        / np.sqrt(
                            xcart[0, :, 0] ** 2
                            + ycart[0, :, 0] ** 2
                            + zcart[0, :, 0] ** 2
                        )
                    )

                s = fid.readline()  # CELL_DATA (NX-1)(NY-1)(NZ-1)
                slist = s.split()
                data_type = str(slist[0], "utf-8")
                if data_type != "CELL_DATA":  # pragma: no cover
                    fid.close()
                    raise ValueError(
                        f"Expected 'CELL DATA' as produced by PLUTO. "
                        f"Got {data_type!r}"
                    )
                s = fid.readline()  # Line feed

                if meta["cell"] == "edges":
                    if n1 > 1:
                        n1 -= 1
                        V["x1"] = r
                    else:
                        V["x1"] = r
                    if n2 > 1:
                        n2 -= 1
                        V["x2"] = theta
                    else:
                        V["x2"] = theta
                    if n3 > 1:
                        n3 -= 1
                        V["x3"] = phi
                    else:
                        V["x3"] = phi

                # Perform averaging on coordinate system to get cell centers
                # The file contains face coordinates, so we extrapolate to get the cell center coordinates.
                elif meta["cell"] == "centers":
                    if n1 > 1:
                        n1 -= 1
                        V["x1"] = 0.5 * (r[1:] + r[:-1])
                    else:
                        V["x1"] = r
                    if n2 > 1:
                        n2 -= 1
                        V["x2"] = 0.5 * (theta[1:] + theta[:-1])
                    else:
                        V["x2"] = theta
                    if n3 > 1:
                        n3 -= 1
                        V["x3"] = 0.5 * (phi[1:] + phi[:-1])
                    else:
                        V["x3"] = phi
            else:
                assert_never(V["geometry"])
        else:
            assert_never(V["geometry"])

        if meta["computedata"]:
            new_shape = n3, n2, n1
            grid_size = int(np.prod(new_shape))
            while 1:
                # SCALARS/VECTORS name data_type (ex: SCALARS imagedata unsigned_char)
                s = fid.readline()
                if len(s) < 2:  # leave if end of file
                    break
                slist = s.split()
                datatype = str(slist[0], "utf-8")
                varname = str(slist[1], "utf-8").upper()
                if datatype == "SCALARS":
                    fid.readline()  # LOOKUP TABLE

                    # we store the file pointer position before computing points
                    inipos = fid.tell()

                    # some smart memory efficient way to store the array
                    array = np.memmap(
                        fid, mode="r", dtype=dt, offset=inipos, shape=grid_size
                    ).reshape(new_shape)
                    # we calculate the offset that we would expect normally with a np.fromfile
                    newpos = inipos + dt.itemsize * grid_size
                    # we set the file pointer position to this offset
                    fid.seek(newpos, os.SEEK_SET)

                    V["data"][varname] = np.transpose(array)
                elif datatype == "VECTORS":
                    # we store the file pointer position before computing points
                    inipos = fid.tell()
                    nelements = 3 * grid_size
                    # some smart memory efficient way to store the array
                    Q = np.memmap(
                        fid, mode="r", dtype=dt, offset=inipos, shape=nelements
                    )
                    # we calculate the offset that we would expect normally with a np.fromfile
                    newpos = inipos + dt.itemsize * nelements
                    # we set the file pointer position to this offset
                    fid.seek(newpos, os.SEEK_SET)

                    for i, suffix in enumerate("XYZ"):
                        name = f"{varname}_{suffix}"
                        V["data"][name] = np.transpose(Q[i::3].reshape(new_shape))

                else:  # pragma: no cover
                    fid.close()
                    raise ValueError(
                        f"Unknown datatype {datatype!r}. "
                        "Expected 'SCALARS' or 'VECTORS'"
                    )

                fid.readline()  # extra line feed
        fid.close()

        return BinData(**V)


class _FargoReader(ReaderMixin, ABC):
    @staticmethod
    def parse_output_number_and_filename(
        file_or_number: Union[PathT, int],
        *,
        directory: PathT,
        prefix: str,  # noqa ARG004
    ) -> tuple[int, Path]:
        directory = Path(directory).resolve()
        if isinstance(file_or_number, int):
            output_number = file_or_number
            file = directory / f"gasdens{output_number:04d}.dat"
        else:
            file = Path(file_or_number)
            if file == Path(file.name):
                file = directory / file
            if len(matches := re.findall(r"\d+", file.name)) == 1:
                output_number = int(matches[0])
            elif len(matches) == 0:
                raise RuntimeError(
                    rf"Failed to guess an output number from {file_or_number!r}"
                )
            else:
                raise RuntimeError(rf"Ambiguous output number from {file_or_number!r}")
        return output_number, file

    @staticmethod
    def get_bin_files(directory: PathT, /) -> list[Path]:
        directory = Path(directory)
        return [
            fn
            for fn in sorted(directory.glob("gasdens*.dat"))
            if re.search(r"gasdens\d+.dat$", str(fn)) is not None
        ]

    @staticmethod
    def _get_output_number_and_dir_from(file) -> tuple[int, Path]:
        _in_file = Path(file).resolve()
        directory = _in_file.parent
        if (match := re.search(r"(?P<on>\d+).dat$", _in_file.name)) is not None:
            output_number = int(match.group("on"))
        else:
            raise ValueError(f"Failed to parse filename {file!r}")

        return output_number, directory

    @staticmethod
    @abstractmethod
    def read(file: PathT, /, **meta) -> BinData: ...


@final
class Fargo3DReader(_FargoReader):
    @override
    @staticmethod
    def read(
        file: PathT,
        /,
        **meta,
    ) -> BinData:
        output_number, directory = _FargoReader._get_output_number_and_dir_from(file)

        default_fluid = "gas"
        fluid_option: Optional[str] = meta.get("fluid", default_fluid)
        if fluid_option is None:
            fluid = default_fluid
        else:
            fluid = fluid_option

        V = BinData.default_init()
        geometry_str = meta["COORDINATES"]
        V["data"] = {}

        domain_x = np.loadtxt(directory / "domain_x.dat")
        # We avoid ghost cells
        domain_y = np.loadtxt(directory / "domain_y.dat")[3:-3]
        domain_z = np.loadtxt(directory / "domain_z.dat")
        if domain_z.shape[0] > 6:
            domain_z = domain_z[3:-3]

        V["x1"] = domain_y  # X-Edge
        if geometry_str == "cylindrical":
            V["geometry"] = Geometry.POLAR
            V["x2"] = domain_x  # Y-Edge
            V["x3"] = domain_z  # Z-Edge #latitute
            pairs = [("RHO", "dens"), ("VX1", "vy"), ("VX2", "vx"), ("VX3", "vz")]
        elif geometry_str == "spherical":
            V["geometry"] = Geometry.SPHERICAL
            V["x2"] = domain_z  # Z-Edge #latitute
            V["x3"] = domain_x  # Y-Edge
            pairs = [("RHO", "dens"), ("VX1", "vy"), ("VX2", "vz"), ("VX3", "vx")]
        else:
            raise NotImplementedError(f"Geometry {geometry_str!r} is not supported")

        n1 = len(V["x1"]) - 1
        n2 = len(V["x2"]) - 1
        n3 = len(V["x3"]) - 1
        grid_shape = n3, n1, n2
        shift = n2 // 2

        def _read_array(file: Path):
            return np.roll(
                np.fromfile(file, dtype="float64")
                .reshape(grid_shape)
                .transpose(1, 2, 0),
                shift,
                axis=1,
            )

        for key, field in pairs:
            file = directory / f"{fluid}{field}{output_number}.dat"
            if not file.is_file():
                continue
            V["data"][key] = _read_array(file)

        if not V["data"]:
            raise FileNotFoundError(
                f"No file matches the pattern '{fluid}*{output_number}.dat'"
            )

        return BinData(**V)


@final
class FargoADSGReader(_FargoReader):
    @override
    @staticmethod
    def read(
        file: PathT,
        /,
        **meta,  # noqa: ARG004
    ) -> BinData:
        output_number, directory = _FargoReader._get_output_number_and_dir_from(file)

        V = BinData.default_init()
        V["geometry"] = Geometry.POLAR
        V["data"] = {}

        phi = np.loadtxt(directory / "used_azi.dat")[:, 0]
        domain_x = np.zeros(len(phi) + 1)
        domain_x[:-1] = phi
        domain_x[-1] = 2 * np.pi
        domain_x -= np.pi
        # We avoid ghost cells
        domain_y = np.loadtxt(directory / "used_rad.dat")
        domain_z = np.zeros(2)

        V["x1"] = domain_y  # X-Edge
        V["x2"] = domain_x  # Y-Edge
        V["x3"] = domain_z  # Z-Edge #latitute

        n1 = len(V["x1"]) - 1
        n2 = len(V["x2"]) - 1
        n3 = len(V["x3"]) - 1
        grid_shape = n3, n1, n2

        def _read_array(file: Path):
            return (
                np.fromfile(file, dtype="float64")
                .reshape(grid_shape)
                .transpose(1, 2, 0)
            )

        for key, field in [("RHO", "dens"), ("VX1", "vrad"), ("VX2", "vtheta")]:
            file = directory / f"gas{field}{output_number}.dat"
            if not file.is_file():
                continue
            V["data"][key] = _read_array(file)
        return BinData(**V)


class NPYReader(ReaderMixin):
    # we accept a leading '_' for backward compatibility
    _filename_re = re.compile(
        r"^_?(?P<prefix>[\w\.]*)"
        r"_(?P<field_name>[A-Z\d]+)"
        r"\.(?P<output_number>\d+)"
        r"\.npy$"
    )

    @staticmethod
    def parse_output_number_and_filename(
        file_or_number: Union[PathT, int],
        *,
        directory: PathT,
        prefix: str,
    ) -> tuple[int, Path]:
        directory = Path(directory).resolve()
        if isinstance(file_or_number, (str, Path)):
            file = Path(file_or_number)
            if (match := NPYReader._filename_re.fullmatch(file.name)) is None:
                raise ValueError(f"Filename {file.name!r} is not recognized")
            if file == Path(file.name):
                file = directory / match.group("field_name").lower() / file
            output_number = int(match.group("output_number"))
            file_alt = None
        else:
            output_number = file_or_number
            all_bin_files = NPYReader.get_bin_files(directory / "any")
            _filter_re = re.compile(rf"^_?{prefix}_[A-Z\d]+.{output_number:04d}.npy")
            matches = [
                file for file in all_bin_files if _filter_re.fullmatch(file.name)
            ]
            if not matches:
                raise FileNotFoundError(
                    "Failed to locate a file matching "
                    f"{prefix=!r} and {output_number=}"
                )
            file = matches[0]
            file_alt = (
                None if file.name.startswith("_") else file.with_name(f"_{file.name}")
            )

        if not file.is_file():
            if file_alt is None:
                raise FileNotFoundError(str(file))

            if not file_alt.is_file():
                raise FileNotFoundError(f"{file} (also tried {file_alt})")
            # backward compatibility
            file = file_alt

        return output_number, file

    @staticmethod
    def get_bin_files(directory: PathT, /) -> list[Path]:
        # return *all* loadable files
        # (not just the ones matching a particular prefix)
        directory = Path(directory).resolve()
        file_paths: list[Path] = []
        for subdir in directory.parent.glob("*"):
            if not subdir.is_dir():
                continue
            if not re.fullmatch(r"\w+", subdir.name):
                continue
            for file in subdir.glob("*.npy"):
                if not file.is_file():
                    continue
                if NPYReader._filename_re.fullmatch(file.name) is None:
                    continue
                file_paths.append(file)

        return sorted(file_paths)

    @staticmethod
    def read(file: PathT, /, **meta) -> BinData:
        meta.setdefault("prefix", "")

        ref_file = Path(file).resolve()
        if (match := NPYReader._filename_re.fullmatch(ref_file.name)) is None:
            raise ValueError(f"Filename {ref_file.name!r} is not recognized")

        prefix = match.group("prefix")
        output_number = int(match.group("output_number"))

        op_suffix = f"_{prefix}" if prefix else ""
        header = ref_file.parents[1] / "header" / f"header{op_suffix}.json"
        with open(header) as fh:
            header_data = json.load(fh)

        geometry = Geometry(header_data.pop("geometry"))
        coordinates: dict[str, FloatArray] = {
            k: np.array(v, dtype="float32") for k, v in header_data.items()
        }
        x1, x2, x3 = coordinates.values()
        n1, n2, n3 = (len(x) for x in (x1, x2, x3))

        # field discovery
        fields_found: dict[str, Path] = {}
        haystack = NPYReader.get_bin_files(ref_file.parent)
        for file in haystack:
            match = NPYReader._filename_re.fullmatch(file.name)
            if match is None:  # pragma: no cover
                raise RuntimeError
            if match.group("prefix") != prefix:
                continue
            if int(match.group("output_number")) != output_number:
                continue
            fields_found[match.group("field_name")] = file

        # sanity check: we should have rediscovered our starting file by now
        assert ref_file in fields_found.values()

        data: dict[str, FloatArray] = {
            k: np.load(v, allow_pickle=True) for k, v in fields_found.items()
        }

        return BinData(data, geometry, x1, x2, x3)
